#include "omp.h"
#include "time.h"
#include <math.h>
#include <vector>
#include <iostream>
#include <fstream>
#include <sstream>
#include "mex.h"

#include "Ensemble.h"	

using namespace std;

#define PI (3.141592653589793)

Ensemble::Ensemble() {
	//mexPrintf("created ensemble %x", this);
	//std::cout << "created ensemble " << this << std::endl;
}

Ensemble::Ensemble(EnsembleParams params_) {
	//mexPrintf("created ensemble %x", this);
	//std::cout << "created ensemble " << this << std::endl;
    params = params_;
}


Ensemble::~Ensemble() {
	for (size_t t = 0; t < trees.size(); t++) {
		delete trees[t];
	}
}

void Ensemble::train(Matrix2df& X, Matrix2df& Y) {
  
    srand(unsigned(time(NULL)));
    
    vector<int> seeds;
	seeds.reserve(params.numTrees);
	for (int t = 0; t < params.numTrees; ++t) {
		seeds.push_back(rand());
	}

	float percentComp = 0;

	// allocate the trees
	trees.resize(params.numTrees);
    
    // count the training data
    vector < vector <int> > classCnt;	
    countTrainingData(X, Y, classCnt);
    
    // create row major difference matrix 
    // speeds up cost sensitive calculation
    if ( params.nodeType == "costsensitive" )
        Y.createDiffMatrix();
    
	#pragma omp parallel for
	for (int eIt=0; eIt< params.numTrees; eIt++) {
		//cout << omp_get_thread_num()<< " " << rand() << endl;
		srand(seeds[eIt]);
		
		//Show progress
		//cout << "\r" << (percentComp/(float)params.numTrees)*100 << "% complete";
		
		// initialise new tree and train
		trees[eIt] = new Tree(params);
		trees[eIt]->train(X, Y, classCnt);
		percentComp ++;
	}
    
    //mexPrintf("nodes %d\n", (int)countNumNonLeafNodes());    
}


void Ensemble::countTrainingData(Matrix2df& X, Matrix2df& Y, vector < vector <int> >& classCnt) {
       
    // want to keep track of the amount of examples in each class. Only doing this if we are using the equal replacement bagging
    if (params.baggingType==3) {
        // initialize count vector
        for (int cIt = 0; cIt < params.noClasses; cIt++)
            classCnt.push_back(vector <int> ());

        // put the examples into different bins depending on class label
        for (int n = 0; n < X.rows; n++) {            
            
            int classId = (int)Y(n,0);
			
			// if Y has more than 1 col we need to find the maximum class
            // TODO speed this up by computing it outside. All the trees will be doign this. 
            // Could just pass the vector of vectors: classCnt
            if( Y.cols > 1) {
                classId = rand()%params.noClasses;
                float maxClassVal = Y(n,classId);
                for (int cIt = 0; cIt < params.noClasses; cIt++) {
                    if (Y(n,cIt) > maxClassVal) {
                        classId = cIt;
						maxClassVal = Y(n,cIt);
					}
                }
            }
            classCnt[classId].push_back(n);            
        }
    }
}


void Ensemble::test(Matrix2df& Xtest, Matrix2df& op) {

	vector<Matrix2df> res(trees.size()); 

	for (size_t tIt = 0; tIt < trees.size(); tIt ++) {
		// initialse matrix num exs X num classes
		res[tIt].create(Xtest.rows, params.noClasses); 		
	}

	#pragma omp parallel for
	for (int tIt = 0; tIt < (int)trees.size(); tIt ++) {
		trees[tIt]->test(Xtest,res[tIt]);
	}	
	
	// getting average of all the classes
	//op.create(Xtest.rows, params.noClasses, 0.0);
	for (size_t tIt = 0; tIt < trees.size(); tIt ++) {
		//op = op + res[tIt];
		op += res[tIt];
	}


	//op.normAcrossRow();
	op = op / (float)trees.size();
	
}

void Ensemble::testSpecifyDepth(Matrix2df& Xtest, Matrix2df& op, int desiredDepth) {

	vector<Matrix2df> res(trees.size()); 

	for (size_t tIt = 0; tIt < trees.size(); tIt ++) {
		// initialse matrix num exs X num classes
		res[tIt].create(Xtest.rows, params.noClasses); 		
	}

	#pragma omp parallel for
	for (int tIt = 0; tIt < (int)trees.size(); tIt ++) {
		trees[tIt]->testSpecifyDepth(Xtest,res[tIt], desiredDepth);
	}	
	
	// getting average of all the classes
	for (size_t tIt = 0; tIt < trees.size(); tIt ++) {
		op += res[tIt];
	}

	op = op / (float)trees.size();
}


double Ensemble::countNumNonLeafNodes() {
	// op the num of non leaf nodes
	double noDecNodes = 0;
	for (size_t tIt = 0; tIt < trees.size(); tIt ++) {
		noDecNodes += (trees[tIt]->noNodesInTree - trees[tIt]->numLeafNodes);
	}	
	return noDecNodes;
}

double Ensemble::countNumNodes() {
	// op the num of total nodes in the forest
	double noDecNodes = 0;
	for (size_t tIt = 0; tIt < trees.size(); tIt ++) {
		noDecNodes += trees[tIt]->noNodesInTree;
	}	
	return noDecNodes;
}

// write forest
void Ensemble::writeEnsemble(string fileName) {
	
	ofstream opFile;
	opFile.open(fileName);	
	
	opFile << params.noClasses << " NumClasses"  << endl;
	opFile << params.numTrees << " NumTrees"  << endl;
	opFile << params.noDims  << " Dimensionality" << endl;
	opFile << params.nodeType  << " NodeType" << endl;

	for (size_t tIt = 0; tIt < trees.size(); tIt ++) {
    	opFile << "Tree " << tIt << endl;		
		opFile << trees[tIt]->noNodesInTree << " Nodes" << endl;
		trees[tIt]->writeTree(trees[tIt]->root, opFile);
	}	
	opFile.close();
}

// TODO
void Ensemble::loadEnsemble(string fileName) {

}