#include "Tree.h"	
#include <algorithm>
#include <math.h>
#include <time.h>
#include <vector>
#include <string>
#include <fstream>
#include <iostream>
//#include "mex.h"

using namespace std;


Tree::Tree(EnsembleParams params_) {

	params = params_;
		
    // problem here if depth is to big, leafNodes overflows as its an int
	leafNodes = (int)pow(2.0, params.treeDepth+1) - 2; 	
	
	// creates new node, type dependent on params, returns if there is no node of type
	root = createNode(0, params);	
	noNodesInTree = 1;
	numLeafNodes = 0;
}


bool Tree::optimizeNode(Matrix2df& X, Matrix2df& Y, Node* curNode) {
	
	// do unary checks at node, check if it can be further split, check if its to deep. if it does not pass - do not split
	if(curNode->impurity==0 || curNode->nodeId >= pow(2.0, params.treeDepth)-1) {
		curNode->setAsLeaf(false);
		numLeafNodes++;
		return false;
	}
		
	curNode->createChildren();
    
	bool betterThanWorstSplit = false;
	//bool bestInfoGainFound = false;
	double bestInfoGain = -FLT_MAX;
    int bestDim;
    float bestThresh;
    
    
	int dIt=0;
    //while (dIt < params.numRandomTests && !bestInfoGainFound) {		
	while (dIt < params.numRandomTests) {		
		
		// deleting and creating new test object every time, not ideal if the test type will be the same for all nodes!
        curNode->getFeatureTest(); // not deleting these properly		
		curNode->computeMinMax(X, Y);
        
		int tIt = 0;
        //while (tIt < params.numThreshTests && !bestInfoGainFound) {
		while (tIt < params.numThreshTests) {
            
            // Set random threshold within min and max distance and evaluate which sides each example go
			curNode->setRandThresh();			
			curNode->train(X,Y); // rename 

			// Calculate info gain and set bestTest pointer if better	
            bool enoughExsAtChildren = ((curNode->left->exsAtNode >= params.minNoOFExsAtNode ) && (curNode->right->exsAtNode >= params.minNoOFExsAtNode) );            
			if (enoughExsAtChildren && (curNode->calcNodeInfoGain(X, Y) > bestInfoGain)) {
            //if (curNode->calcNodeInfoGain(X, Y) > bestInfoGain) {

				bestDim = curNode->dimId;
                bestThresh = curNode->thresh;
                
                bestInfoGain = curNode->infoGain;	
				betterThanWorstSplit = true;
                
                // Exit if best possible split found
                //if (curNode->infoGain == curNode->impurity) {
                //  bestInfoGainFound = true; 
                //}
			}
                       
			tIt++;
        }
		dIt++;
    }
    

	// copy best computed test
	// recompute child nodes and indices for them	
	if (betterThanWorstSplit) {
		
        curNode->dimId = bestDim;
        curNode->thresh = bestThresh;
		curNode->updateLeftRight(X,Y);
        
        curNode->calcNodeInfoGain(X, Y);   
        
        curNode->normalizePosterior();
        
		bool enoughExsAtChildren = ((curNode->left->exsAtNode >= params.minNoOFExsAtNode ) && (curNode->right->exsAtNode >= params.minNoOFExsAtNode) );			
		if (betterThanWorstSplit && enoughExsAtChildren) {	
			noNodesInTree += 2; 
			return true;
		}

	} 
	
	// do not split and set node as leaf
	curNode->setAsLeaf(true);
	numLeafNodes++;
	return false; 	
}

void Tree::buildTree(Matrix2df& X, Matrix2df& Y, Node* curNode) {

	bool successfullSplit = optimizeNode(X, Y, curNode);

	if (successfullSplit) {
		buildTree(X, Y, curNode->left); 
		buildTree(X, Y, curNode->right); 
	}
}


void Tree::train(Matrix2df& X, Matrix2df& Y, vector<vector <int>>& classCnt) {

    // do bagging - and initialize root
	root->doBagging(X, Y, classCnt);
	root->initializeRoot(X, Y);
	
    buildTree(X, Y, root);

}


// traverse tree and add node in correct position
void Tree::addNodeToTree(Node* node, Node* nodeToAdd) {

	if (2*(node->nodeId)+1 == nodeToAdd->nodeId) {
		node->left = nodeToAdd;
		node->isLeaf = false;
	}
	else if (2*(node->nodeId)+2 == nodeToAdd->nodeId) {
		node->right = nodeToAdd;
		node->isLeaf = false;
	}
	else {
		if (node->left !=0)	
			addNodeToTree(node->left, nodeToAdd);
		if (node->right !=0)	
			addNodeToTree(node->right, nodeToAdd);
	}
}

void Tree::addNode(Node* node) {

	node->isLeaf = true; 
	if (node->nodeId == 0) {
		root = node;
	}
	else
		addNodeToTree(root, node);
}


// return all the posteriors
int Tree::test(Matrix2df& X, Matrix2df& res) {
	
	for (int xIt = 0; xIt < X.rows; xIt++) {		
		Node* node = root;		

		while (!node->isLeaf) { // exit if it hits a leaf	

			// test, 0 to go left and 1 to go right
			if (node->test(X, xIt)) 
				node = node->right;
			else 
				node = node->left;	
		}
		node->opPosterior(res, xIt) ;				
	}
	return(0);
}


// only test until a specified depth
int Tree::testSpecifyDepth(Matrix2df& X, Matrix2df& res, int desiredDepth) {
	    
	for (int xIt = 0; xIt < X.rows; xIt++) {		
		Node* node = root;		
        int curDepth = 0;
		while (!node->isLeaf) { // exit if it hits a leaf	

            // escape wen we reach the desired depth
            if (curDepth >= desiredDepth)
                break;
                
			// test, 0 to go left and 1 to go right
			if (node->test(X, xIt)) 
				node = node->right;
			else 
				node = node->left;	
            
            curDepth++;
		}
		node->opPosterior(res, xIt) ;				
	}
	return(0);
}


void Tree::writeTree(Node* node, ofstream& opFile) {

	// node to string
	opFile << node->nodeToString();
	opFile << "\n";

	// traverse the tree
	if (!node->isLeaf) {		
		writeTree(node->left, opFile);	
		writeTree(node->right, opFile);	
	}
}

Tree::~Tree() {
	delete root;
}