#ifndef UTILS_H
#define UTILS_H
#include <stdlib.h>
#include <string>
#include <vector>
//#include "mex.h"

class EnsembleParams{
public:
	std::string ipData;	
	std::string ipLabels;
	std::string ensembleFile;
	std::string testData;
	std::string opFileName;

	int numTrees;
	int treeDepth; // 0 = 1 node, 1 = 3 nodes, 2 = 7 ...
	int numRandomTests;//num of random tests at each node
	int numThreshTests;//num of random thresholds at that node
	float bagProb; // probability of sample landing in bag
	int minNoOFExsAtNode; // stop growing tree if num of examples at node equals this value
	int baggingType; //0 bagging with rep, 1 bagging no rep (give bagProb), 2 no bagging

	int noDims;
	int noClasses;

	// specify impurity measure to use
	std::string nodeType; // options: classification, costsensitive, costsensitivenaive
	};


// TODO add exception handling 
class Matrix2df{

public:
	int rows;
	int cols;
	float* data;
    double* dataDiff;
	bool fromML;
    bool hasDiffMatrix;
    int diffMatrixCols;
    
	Matrix2df() {
		rows = 0;
		cols = 0;
		fromML = false;
        hasDiffMatrix = false;
	}

	
	Matrix2df(int rows_, int cols_, float init) {
		rows  = rows_;
		cols = cols_;
		fromML = false;        
        hasDiffMatrix = false;
		data = new float[rows*cols];		
		for (int i = 0; i < rows*cols; ++i) {
			data[i] = init;
		}
	}


	Matrix2df(int rows_, int cols_, float* dataToPt) {
		rows  = rows_;
		cols = cols_;
		data = dataToPt;
		fromML = true;        
        hasDiffMatrix = false;
	}

	
	Matrix2df(int rows_, int cols_) {
		rows  = rows_;
		cols = cols_;
		fromML = false;        
        hasDiffMatrix = false;
		data = new float[rows*cols];	
	}
	

	// copy constructor
	Matrix2df(const Matrix2df& m) {
		rows = m.rows;
		cols = m.cols;
		fromML = m.fromML;
        hasDiffMatrix = m.hasDiffMatrix;
        
		data = new float[rows*cols];		
		for (int i = 0; i < rows*cols; ++i) {
			data[i] = m.data[i];
		}
	}

	void create(int rows_, int cols_) {
		rows  = rows_;
		cols = cols_;
		fromML = false;
        hasDiffMatrix = false;        
		data = new float[rows*cols];	
	}

	void create(int rows_, int cols_, float init) {
		rows  = rows_;
		cols = cols_;        
		fromML = false;
        hasDiffMatrix = false;  
		data = new float[rows*cols];	
		for (int i = 0; i < rows*cols; ++i) {
			data[i] = init;
		}
	}

    
	void setToZero() {
		for (int i = 0; i < rows*cols; i++) {			
				data[i] = 0;
		}
	}


	void normAcrossRow() {

		for (int rIt = 0; rIt < rows; rIt ++) {
			float norm = 0;
			for (int cIt = 0; cIt < cols; cIt ++) {
				norm+=data[rIt + cIt*rows];
			}
			for (int cIt = 0; cIt < cols; cIt ++) {
				data[rIt + cIt*rows] = data[rIt + cIt*rows]/norm;
			}
		}
	}

	void copyRow(const Matrix2df& m, int rowId){
		for (int i=0; i< cols;i++)
			data[i] = m.data[rowId + i*rows];		
	}
	

	Matrix2df& operator+=(const Matrix2df& m) {
		for (int i = 0; i < rows*cols; i++)
			data[i] = data[i]+m.data[i];
		
		return *this;
	}

	

	Matrix2df operator+ (const Matrix2df& m1) {
		Matrix2df res(*this);
		return res+=m1;
	}


	Matrix2df& operator = (const Matrix2df& m1) {

		if (m1.cols == cols && m1.rows == rows) {
			for (int i = 0; i< m1.rows*m1.cols; i++)
					data[i] = m1.data[i];
		}
		return(*this);
	}

	Matrix2df& operator/ (float denom) {

		for (int i = 0; i < rows*cols; ++i)
			data[i] = data[i] / float(denom);

		return(*this);
	}

	// column major order
	float &Matrix2df::operator()(int row, int col) {
		return data[row + col*rows];
	}
    
    
    
     // makes a deep copy and creates a diff matrix for cost sensitive learning
     // note that its also in row order
    void createDiffMatrix() {
        
        hasDiffMatrix = true;
        diffMatrixCols = (cols*cols - cols);        
        
        // dataDiff does not store diagonal data
        dataDiff = new double[rows*diffMatrixCols];
        for (int d=0; d< rows*diffMatrixCols; d++)
            dataDiff[d] = 0;        
        
        for (int r = 0; r < rows; r++) {
            
            // just indexing upper triangular matrix
            for (int i=0; i < cols;i++) {
                for (int j=i+1; j < cols;j++) {
                
                        double Yi = (double)data[r + i*rows];
                        double Yj = (double)data[r + j*rows];
                        double d_ij = 0.0;
                        int ind = 0;
                    
                        if (Yi > Yj) {
                            d_ij = Yi - Yj;                            
                            ind = i+j*cols - (j + (j<i)); 
                        } else {
                            d_ij = Yj - Yi;
                            ind = j+i*cols - (i + (i<j));                            
                        }
                        
                        // there will be a 0 if its not greater than
                        dataDiff[ind+r*diffMatrixCols] = d_ij;
                }
            } 
        }
        
    }
    
    
    float diffMatrix(int row, int col) {
        return dataDiff[col + row*diffMatrixCols];    
    }
    

	~Matrix2df() {	

		if (cols !=0 && rows !=0 && !fromML) {
			delete [] data;            
		}
        if (hasDiffMatrix) {
            delete [] dataDiff;
        }
	}

};

#endif